# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-03-19 14:46
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0018_auto_20180316_1305")]

    operations = [
        migrations.CreateModel(
            name="ArchivedOutbounds",
            fields=[
                (
                    "date",
                    models.DateField(
                        help_text=b"The date that the archive is for",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "archive",
                    models.FileField(
                        help_text=b"The file for the archive", upload_to=b""
                    ),
                ),
            ],
            options={
                "verbose_name": "archived outbounds",
                "verbose_name_plural": "archived outbounds",
            },
        ),
        migrations.AlterModelOptions(
            name="aggregateoutbounds",
            options={
                "verbose_name": "aggregate outbounds",
                "verbose_name_plural": "aggregate outbounds",
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-05-03 12:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0013_outbound_channel")]

    operations = [
        migrations.AddField(
            model_name="outbound",
            name="to_identity",
            field=models.CharField(
                blank=True, max_length=36, null=False, db_index=True
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="to_addr",
            field=models.CharField(
                blank=True, db_index=True, max_length=500, null=False
            ),
        ),
        migrations.AddField(
            model_name="inbound",
            name="from_identity",
            field=models.CharField(
                blank=True, max_length=36, null=False, db_index=True, default=""
            ),
        ),
        migrations.AlterField(
            model_name="inbound",
            name="from_addr",
            field=models.CharField(
                blank=True, db_index=True, max_length=255, null=False
            ),
        ),
    ]



import re

import phonenumbers


def noop(msisdn):
    return msisdn


def vas2nets_voice(msisdn):
    """
    FIXME: this should not need be in this repo

    Vas2Nets is an aggregator in Nigeria, for some reason they need
    MSISDNs prefixed with a 9 instead of the country code to initiate an OBD.
    """
    return re.sub(r"\+?234(\d+)$", r"90\1", msisdn)


def vas2nets_text(msisdn):
    """
    FIXME: this should not need be in this repo

    Vas2Nets is an aggregator in Nigeria, they need MSISDNs in the local
    format, prefixed with a 0, instead of the international format with the
    country code.
    """
    return re.sub(r"\+234(\d+)$", r"234\1", msisdn)


def e_164(msisdn: str) -> str:
    """
    Returns the msisdn in E.164 international format.
    """
    # Phonenumbers library requires the + to identify the country, so we add it if it
    # does not already exist
    number = phonenumbers.parse("+{}".format(msisdn.lstrip("+")), None)
    return phonenumbers.format_number(number, phonenumbers.PhoneNumberFormat.E164)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-11-25 13:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0003_auto_20161124_1357")]

    operations = [
        migrations.AddField(
            model_name="outbound",
            name="last_sent_time",
            field=models.DateTimeField(blank=True, null=True),
        )
    ]



"""
Django settings for seed_message_sender project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

import os

import dj_database_url

from kombu import Exchange, Queue
from getenv import env

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env("SECRET_KEY", "REPLACEME")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env("DEBUG", False)

ALLOWED_HOSTS = ["*"]


# Application definition

INSTALLED_APPS = (
    # admin
    "django.contrib.admin",
    # core
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    # 3rd party
    "raven.contrib.django.raven_compat",
    "rest_framework",
    "rest_framework.authtoken",
    "django_filters",
    "rest_hooks",
    "storages",
    "django_prometheus",
    # us
    "message_sender",
)

SITE_ID = 1
USE_SSL = os.environ.get("USE_SSL", "false").lower() == "true"
USE_SSL = env("USE_SSL", False)

MIDDLEWARE = (
    "django_prometheus.middleware.PrometheusBeforeMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_prometheus.middleware.PrometheusAfterMiddleware",
)

ROOT_URLCONF = "seed_message_sender.urls"

WSGI_APPLICATION = "seed_message_sender.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    "default": dj_database_url.config(
        default=env(
            "MESSAGE_SENDER_DATABASE",
            "postgres://postgres:@localhost/seed_message_sender",
        ),
        engine="django_prometheus.db.backends.postgresql",
    )
}

PROMETHEUS_EXPORT_MIGRATIONS = False


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = "en-gb"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "django.contrib.staticfiles.finders.FileSystemFinder",
)

STATIC_ROOT = "static"
STATIC_URL = "/static/"

MEDIA_ROOT = "media"
MEDIA_URL = "/media/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

# Sentry configuration
RAVEN_CONFIG = {
    # DevOps will supply you with this.
    "dsn": env("MESSAGE_SENDER_SENTRY_DSN", None)
}

# REST Framework conf defaults
REST_FRAMEWORK = {
    "PAGE_SIZE": 1000,
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.CursorPagination",
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.BasicAuthentication",
        "seed_message_sender.auth.CachedTokenAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
}

# Webhook event definition
HOOK_EVENTS = {
    # 'any.event.name': 'App.Model.Action' (created/updated/deleted)
    # 'dummymodel.added': 'message_sender.DummyModel.created+'
    "outbound.delivery_report": None,
    "whatsapp.failed_contact_check": None,
    "identity.no_address": None,
}

HOOK_DELIVERER = "message_sender.tasks.deliver_hook_wrapper"

HOOK_AUTH_TOKEN = env("HOOK_AUTH_TOKEN", "REPLACEME")

CELERY_BROKER_URL = env("BROKER_URL", "redis://localhost:6379/0")

CELERY_TASK_DEFAULT_QUEUE = "seed_message_sender"
CELERY_TASK_QUEUES = (
    Queue(
        "seed_message_sender",
        Exchange("seed_message_sender"),
        routing_key="seed_message_sender",
    ),
)

CELERY_TASK_ALWAYS_EAGER = False

# Tell Celery where to find the tasks
CELERY_IMPORTS = ("message_sender.tasks",)

CELERY_TASK_CREATE_MISSING_QUEUES = True
CELERY_TASK_ROUTES = {
    "celery.backend_cleanup": {"queue": "mediumpriority"},
    "message_sender.tasks.deliver_hook_wrapper": {"queue": "priority"},
    "message_sender.tasks.send_message": {"queue": "lowpriority"},
    "message_sender.tasks.fire_metric": {"queue": "metrics"},
    "message_sender.tasks.requeue_failed_tasks": {"queue": "mediumpriority"},
}

METRICS_REALTIME = [
    "vumimessage.tries.sum",
    "vumimessage.maxretries.sum",
    "vumimessage.obd.tries.sum",
    "message.failures.sum",
    "message.sent.sum",
    "sender.send_message.connection_error.sum",
    "sender.send_message.http_error.400.sum",
    "sender.send_message.http_error.401.sum",
    "sender.send_message.http_error.403.sum",
    "sender.send_message.http_error.404.sum",
    "sender.send_message.http_error.500.sum",
    "sender.send_message.timeout.sum",
]
METRICS_SCHEDULED = []
METRICS_SCHEDULED_TASKS = []

CELERY_TASK_SERIALIZER = "json"
CELERY_RESULT_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = ["json"]
CELERY_TASK_IGNORE_RESULT = True
CELERY_WORKER_MAX_TASKS_PER_CHILD = 50

MESSAGE_BACKEND_VOICE = env("MESSAGE_SENDER_MESSAGE_BACKEND_VOICE", "vumi")
MESSAGE_BACKEND_TEXT = env("MESSAGE_SENDER_MESSAGE_BACKEND_TEXT", "vumi")

VUMI_API_URL_VOICE = env(
    "MESSAGE_SENDER_VUMI_API_URL_VOICE",
    "http://example.com/api/v1/go/http_api_nostream",
)
VUMI_ACCOUNT_KEY_VOICE = env("MESSAGE_SENDER_VUMI_ACCOUNT_KEY_VOICE", "acc-key")
VUMI_CONVERSATION_KEY_VOICE = env(
    "MESSAGE_SENDER_VUMI_CONVERSATION_KEY_VOICE", "conv-key"
)
VUMI_ACCOUNT_TOKEN_VOICE = env("MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_VOICE", "conv-token")

VOICE_TO_ADDR_FORMATTER = env(
    "VOICE_TO_ADDR_FORMATTER", "message_sender.formatters.noop"
)
TEXT_TO_ADDR_FORMATTER = env("TEXT_TO_ADDR_FORMATTER", "message_sender.formatters.noop")

VUMI_API_URL_TEXT = env(
    "MESSAGE_SENDER_VUMI_API_URL_TEXT", "http://example.com/api/v1/go/http_api_nostream"
)
VUMI_ACCOUNT_KEY_TEXT = env("MESSAGE_SENDER_VUMI_ACCOUNT_KEY_TEXT", "acc-key")
VUMI_CONVERSATION_KEY_TEXT = env(
    "MESSAGE_SENDER_VUMI_CONVERSATION_KEY_TEXT", "conv-key"
)
VUMI_ACCOUNT_TOKEN_TEXT = env("MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_TEXT", "conv-token")

JUNEBUG_API_URL_VOICE = env(
    "MESSAGE_SENDER_JUNEBUG_API_URL_VOICE",
    "http://example.com/jb/channels/abc-def/messages",
)
JUNEBUG_API_AUTH_VOICE = env("MESSAGE_SENDER_JUNEBUG_API_AUTH_VOICE", None)
JUNEBUG_API_FROM_VOICE = env("MESSAGE_SENDER_JUNEBUG_API_FROM_VOICE", None)

JUNEBUG_API_URL_TEXT = env(
    "MESSAGE_SENDER_JUNEBUG_API_URL_TEXT",
    "http://example.com/jb/channels/def-abc/messages",
)
JUNEBUG_API_AUTH_TEXT = env("MESSAGE_SENDER_JUNEBUG_API_AUTH_TEXT", None)
JUNEBUG_API_FROM_TEXT = env("MESSAGE_SENDER_JUNEBUG_API_FROM_TEXT", None)

MESSAGE_SENDER_MAX_RETRIES = env("MESSAGE_SENDER_MAX_RETRIES", 3)
MESSAGE_SENDER_MAX_FAILURES = env("MESSAGE_SENDER_MAX_FAILURES", 5)

METRICS_URL = env("METRICS_URL", None)
METRICS_AUTH = (
    env("METRICS_AUTH_USER", "REPLACEME"),
    env("METRICS_AUTH_PASSWORD", "REPLACEME"),
)

REDIS_HOST = env("REDIS_HOST", "localhost")
REDIS_PORT = env("REDIS_PORT", 6379)
REDIS_DB = env("REDIS_DB", 0)

# A value of 0 disables cuncurrency limiter
CONCURRENT_VOICE_LIMIT = env("CONCURRENT_VOICE_LIMIT", 0)
# Seconds to wait before retrying a waiting message
VOICE_MESSAGE_DELAY = env("VOICE_MESSAGE_DELAY", 0)
# Seconds until we assume a message has finished
VOICE_MESSAGE_TIMEOUT = env("VOICE_MESSAGE_TIMEOUT", 0)
# A value of 0 disables cuncurrency limiter
CONCURRENT_TEXT_LIMIT = env("CONCURRENT_TEXT_LIMIT", 0)
# Seconds to wait before retrying a waiting message
TEXT_MESSAGE_DELAY = env("TEXT_MESSAGE_DELAY", 0)
# Seconds until we assume a message has finished
TEXT_MESSAGE_TIMEOUT = env("TEXT_MESSAGE_TIMEOUT", 0)

CACHES = {
    "default": {
        "BACKEND": "django_prometheus.cache.backends.redis.RedisCache",
        "LOCATION": ["%s:%s" % (REDIS_HOST, REDIS_PORT)],
        "OPTIONS": {"DB": REDIS_DB},
    },
    "locmem": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
}
REDIS_PASSWORD = env("REDIS_PASSWORD", None)
if REDIS_PASSWORD:
    CACHES["default"]["OPTIONS"]["PASSWORD"] = REDIS_PASSWORD

DEFAULT_REQUEST_TIMEOUT = env("DEFAULT_REQUEST_TIMEOUT", 30)

IDENTITY_STORE_URL = env("IDENTITY_STORE_URL", "http://is/api/v1")
IDENTITY_STORE_TOKEN = env("IDENTITY_STORE_TOKEN", "REPLACEME")

AGGREGATE_OUTBOUND_BACKTRACK = env("AGGREGATE_OUTBOUND_BACKTRACK", 30)


AWS_ACCESS_KEY_ID = env("AWS_ACCESS_KEY_ID", None)
AWS_SECRET_ACCESS_KEY = env("AWS_SECRET_ACCESS_KEY", None)
AWS_STORAGE_BUCKET_NAME = env("AWS_STORAGE_BUCKET_NAME", None)
AWS_S3_ENCRYPTION = True

if AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY and AWS_STORAGE_BUCKET_NAME:
    DEFAULT_FILE_STORAGE = "storages.backends.s3boto3.S3Boto3Storage"

SAFE_TIME_INTERVAL = env("SAFE_TIME_INTERVAL", None)



from django.contrib import admin
from django import forms
from django.core.exceptions import ValidationError
from django.core.paginator import Paginator
from django.db import connections
from django.utils.functional import cached_property

from .models import Outbound, Inbound, Channel, AggregateOutbounds, ArchivedOutbounds
from .tasks import send_message


class EstimatedCountPaginator(Paginator):
    def __init__(self, *args, **kwargs):
        super(EstimatedCountPaginator, self).__init__(*args, **kwargs)
        self.object_list.count = self.count

    @cached_property
    def count(self):
        if self.object_list.query.where:
            return self.object_list.count()

        db_table = self.object_list.model._meta.db_table
        cursor = connections[self.object_list.db].cursor()
        cursor.execute("SELECT reltuples FROM pg_class WHERE relname = %s", (db_table,))
        result = cursor.fetchone()
        if not result:
            return 0
        return int(result[0])


class OutboundAdmin(admin.ModelAdmin):
    list_display = (
        "to_addr",
        "to_identity",
        "delivered",
        "attempts",
        "vumi_message_id",
        "created_at",
        "updated_at",
        "content",
    )
    list_display_links = ("to_addr", "to_identity")
    list_filter = ("delivered", "attempts", "created_at", "updated_at")
    search_fields = ["to_addr", "to_identity"]
    actions = ["resend_outbound"]
    paginator = EstimatedCountPaginator

    def resend_outbound(self, request, queryset):
        resent = 0
        for record in queryset.iterator():
            send_message.apply_async(kwargs={"message_id": str(record.id)})
            resent += 1
        if resent == 1:
            output_text = "message"
        else:
            output_text = "messages"
        self.message_user(
            request, "Attempting to resend %s %s." % (resent, output_text)
        )

    resend_outbound.short_description = "Resend selected outbounds"


class InboundAdmin(admin.ModelAdmin):
    list_display = (
        "message_id",
        "in_reply_to",
        "to_addr",
        "from_addr",
        "from_identity",
        "created_at",
        "updated_at",
        "content",
    )
    list_filter = ("in_reply_to", "from_addr", "created_at", "updated_at")
    search_fields = ["to_addr", "from_identity"]
    paginator = EstimatedCountPaginator


class ChannelAdminForm(forms.ModelForm):
    def clean(self):
        channel_id = self.cleaned_data["channel_id"]
        channel_type = self.cleaned_data["channel_type"]
        config = self.cleaned_data["configuration"]

        missing = []
        if channel_type == Channel.JUNEBUG_TYPE:
            keys = ("JUNEBUG_API_URL", "JUNEBUG_API_AUTH", "JUNEBUG_API_FROM")

        elif channel_type == Channel.VUMI_TYPE:
            keys = (
                "VUMI_CONVERSATION_KEY",
                "VUMI_ACCOUNT_KEY",
                "VUMI_ACCOUNT_TOKEN",
                "VUMI_API_URL",
            )
        elif channel_type == Channel.HTTP_API_TYPE:
            keys = ("HTTP_API_URL", "HTTP_API_AUTH", "HTTP_API_FROM")
        elif channel_type == Channel.WASSUP_API_TYPE:
            keys = (
                "WASSUP_API_URL",
                "WASSUP_API_TOKEN",
                "WASSUP_API_HSM_UUID",
                "WASSUP_API_NUMBER",
            )
        elif channel_type == Channel.WHATSAPP_API_TYPE:
            keys = ("API_URL", "API_TOKEN")

        for key in keys:
            if key not in config.keys():
                missing.append(key)

        if missing:
            raise ValidationError(
                "Configuration keys missing: {}".format(", ".join(missing))
            )

        if not self.cleaned_data["default"]:
            if (
                not Channel.objects.filter(default=True)
                .exclude(channel_id=channel_id)
                .exists()
            ):
                raise ValidationError("Please make sure there is a default channel.")

        return self.cleaned_data


class ChannelAdmin(admin.ModelAdmin):
    list_display = ("channel_id", "channel_type", "concurrency_limit", "default")
    list_filter = ("channel_type", "default")
    search_fields = ["channel_id"]
    form = ChannelAdminForm


class AggregateOutboundsAdmin(admin.ModelAdmin):
    list_display = ("date", "delivered", "channel", "attempts", "total")
    list_filter = ("date", "delivered", "channel")
    paginator = EstimatedCountPaginator


class ArchivedOutboundsAdmin(admin.ModelAdmin):
    list_display = ("date", "archive")
    list_filter = ("date",)


admin.site.register(Outbound, OutboundAdmin)
admin.site.register(Inbound, InboundAdmin)
admin.site.register(Channel, ChannelAdmin)
admin.site.register(AggregateOutbounds, AggregateOutboundsAdmin)
admin.site.register(ArchivedOutbounds, ArchivedOutboundsAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-16 10:54
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Inbound",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("message_id", models.CharField(max_length=36)),
                ("in_reply_to", models.CharField(blank=True, max_length=36, null=True)),
                ("to_addr", models.CharField(max_length=255)),
                ("from_addr", models.CharField(max_length=255)),
                ("content", models.CharField(blank=True, max_length=1000, null=True)),
                ("transport_name", models.CharField(max_length=200)),
                ("transport_type", models.CharField(max_length=200)),
                ("helper_metadata", django.contrib.postgres.fields.jsonb.JSONField()),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="inbounds_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="inbounds_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Outbound",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("to_addr", models.CharField(max_length=500)),
                ("version", models.IntegerField(default=1)),
                ("content", models.CharField(blank=True, max_length=1000, null=True)),
                (
                    "vumi_message_id",
                    models.CharField(blank=True, max_length=36, null=True),
                ),
                ("delivered", models.BooleanField(default=False)),
                ("attempts", models.IntegerField(default=0)),
                ("metadata", django.contrib.postgres.fields.jsonb.JSONField()),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="outbounds_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="outbounds_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-01 13:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0004_outbound_last_sent_time")]

    operations = [
        migrations.AddField(
            model_name="outbound",
            name="call_answered",
            field=models.NullBooleanField(
                default=None,
                help_text=b"True if the call has been answered. Not used for text messages",
            ),
        )
    ]



from django.conf import settings
from importlib import import_module
from seed_services_client import IdentityStoreApiClient
from datetime import timedelta
from django.utils import timezone
from django.utils.dateparse import parse_datetime

identity_store_client = IdentityStoreApiClient(
    settings.IDENTITY_STORE_TOKEN,
    settings.IDENTITY_STORE_URL,
    retries=5,
    timeout=settings.DEFAULT_REQUEST_TIMEOUT,
)


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    return available_metrics


def load_callable(dotted_path_to_callable):
    module_name, func_name = dotted_path_to_callable.rsplit(".", 1)
    mod = import_module(module_name)
    func = getattr(mod, func_name)
    return func


def get_identity_address(identity_uuid, use_communicate_through=False):
    params = {"default": True}
    if use_communicate_through:
        params["use_communicate_through"] = True

    return identity_store_client.get_identity_address(identity_uuid, params=params)


def get_identity_by_address(address_value, address_type="msisdn"):
    r = identity_store_client.get_identity_by_address(address_type, address_value)

    results = list(r["results"])
    if len(results) > 0:
        return results
    else:
        return None


def create_identity(identity):
    return identity_store_client.create_identity(identity)


def is_in_time_interval(interval, timestamp=None):
    """
    Checks whether a timestamp falls within an interval or not, and when would be the
    next time that it would fall within the interval.

    interval (str): The interval to check, given by two ISO-8601 times,
        `<time1>/<time2>`, eg. "09:00:00Z/17:00:00Z"
    timestamp (datetime, optional): The timestamp to check. Defaults to the current time.

    Returns (in_interval (bool), safe (datetime))
    """
    if timestamp is None:
        timestamp = timezone.now()

    start, end = interval.split("/")
    start = parse_datetime(f"{timestamp.date().isoformat()}T{start}")
    end = parse_datetime(f"{timestamp.date().isoformat()}T{end}")

    in_interval = start < timestamp < end
    if in_interval:
        return in_interval, timestamp
    else:
        if timestamp > end:
            return in_interval, start + timedelta(days=1)
        else:
            return in_interval, start



from datetime import datetime
from unittest.mock import patch

from django.test import TestCase, override_settings

from message_sender.tasks import send_message


class SendMessageTests(TestCase):
    @override_settings(SAFE_TIME_INTERVAL="09:00:00Z/17:00:00Z")
    @patch("message_sender.tasks.send_message.retry")
    @patch("message_sender.tasks.is_in_time_interval")
    def test_retry_outside_of_safe_time(self, is_in_time_interval, retry):
        """
        If we are outside of the safe sending time, then we should retry
        """
        is_in_time_interval.return_value = (False, datetime(2019, 1, 1, 8, 0, 0))
        send_message(None)
        retry.assert_called_once_with(eta=datetime(2019, 1, 1, 8, 0, 0))

    @override_settings(SAFE_TIME_INTERVAL="09:00:00Z/17:00:00Z")
    @patch("message_sender.tasks.send_message.retry")
    @patch("message_sender.tasks.is_in_time_interval")
    def test_no_retry_inside_of_safe_time(self, is_in_time_interval, retry):
        """
        If we are inside the safe sending time, then we should send the message
        """
        is_in_time_interval.return_value = (True, datetime(2019, 1, 1, 8, 0, 0))
        send_message(None)
        retry.assert_not_called()



# Generated by Django 2.1.1 on 2018-09-14 14:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0021_auto_20180504_1446")]

    operations = [
        migrations.AlterField(
            model_name="aggregateoutbounds",
            name="attempts",
            field=models.IntegerField(help_text="The total number of attempts"),
        ),
        migrations.AlterField(
            model_name="aggregateoutbounds",
            name="channel",
            field=models.ForeignKey(
                help_text="Which channel this is for",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="message_sender.Channel",
            ),
        ),
        migrations.AlterField(
            model_name="aggregateoutbounds",
            name="date",
            field=models.DateField(help_text="The date that the aggregate is for"),
        ),
        migrations.AlterField(
            model_name="aggregateoutbounds",
            name="delivered",
            field=models.BooleanField(
                help_text="Whether this is for delivery passed or failed messages"
            ),
        ),
        migrations.AlterField(
            model_name="aggregateoutbounds",
            name="total",
            field=models.IntegerField(help_text="The total number of messages"),
        ),
        migrations.AlterField(
            model_name="archivedoutbounds",
            name="archive",
            field=models.FileField(help_text="The file for the archive", upload_to=""),
        ),
        migrations.AlterField(
            model_name="archivedoutbounds",
            name="date",
            field=models.DateField(
                help_text="The date that the archive is for",
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="channel",
            name="channel_type",
            field=models.CharField(
                choices=[
                    ("junebug", "Junebug"),
                    ("vumi", "Vumi"),
                    ("http_api", "HTTP API"),
                    ("wassup", "Wassup API"),
                ],
                default="junebug",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="inbound",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="inbounds_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="inbound",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="inbounds_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="call_answered",
            field=models.NullBooleanField(
                default=None,
                help_text="True if the call has been answered. Not used for text messages",
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="channel",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="message_sender.Channel",
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="outbounds_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="resend",
            field=models.NullBooleanField(
                default=None,
                help_text="True if this is a resend requested by the user.",
            ),
        ),
        migrations.AlterField(
            model_name="outbound",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="outbounds_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:56
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("message_sender", "0007_add_to_addr_index"),
        ("message_sender", "0009_auto_20170323_1453"),
    ]

    operations = []



import base64
import hmac
from datetime import datetime, timedelta
from hashlib import sha256

from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.shortcuts import get_object_or_404
from django_filters import rest_framework as filters
from prometheus_client import Counter
from rest_framework import mixins, status, viewsets
from rest_framework.authtoken.models import Token
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.filters import OrderingFilter
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_hooks.models import Hook

from seed_message_sender.utils import (
    create_identity,
    get_available_metrics,
    get_identity_by_address,
)

from .formatters import e_164
from .models import (
    AggregateOutbounds,
    ArchivedOutbounds,
    Channel,
    Inbound,
    InvalidMessage,
    Outbound,
    OutboundSendFailure,
)
from .serializers import (
    AggregateOutboundSerializer,
    ArchivedOutboundSerializer,
    CreateUserSerializer,
    EventSerializer,
    HookSerializer,
    InboundSerializer,
    JunebugEventSerializer,
    JunebugInboundSerializer,
    OutboundSendFailureSerializer,
    OutboundSerializer,
    WassupEventSerializer,
    WassupInboundSerializer,
    WhatsAppWebhookSerializer,
    WhatsAppEventSerializer,
    WhatsAppInboundSerializer,
)
from .tasks import (
    ConcurrencyLimiter,
    aggregate_outbounds,
    archive_outbound,
    requeue_failed_tasks,
    send_message,
)

# Uncomment line below if scheduled metrics are added
# from .tasks import scheduled_metrics


class IdCursorPagination(CursorPagination):
    ordering = "-id"


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """

    permission_classes = (IsAdminUser,)

    def post(self, request):
        """Create a user and token, given an email. If user exists just
        provide the token."""
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get("email")
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(status=status.HTTP_201_CREATED, data={"token": token.key})


class HookViewSet(viewsets.ModelViewSet):
    """
    Retrieve, create, update or destroy webhooks.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Hook.objects.all()
    serializer_class = HookSerializer

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class MultipleField(forms.Field):
    widget = forms.MultipleHiddenInput

    def clean(self, value):
        if value is None:
            return None
        return [super(MultipleField, self).clean(v) for v in value]


class MultipleFilter(filters.Filter):
    field_class = MultipleField

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("lookup_expr", "in")
        super(MultipleFilter, self).__init__(*args, **kwargs)


class OutboundFilter(filters.FilterSet):
    before = filters.IsoDateTimeFilter(field_name="created_at", lookup_expr="lte")
    after = filters.IsoDateTimeFilter(field_name="created_at", lookup_expr="gte")
    to_addr = MultipleFilter(field_name="to_addr")
    to_identity = MultipleFilter(field_name="to_identity")

    class Meta:
        model = Outbound
        fields = (
            "version",
            "vumi_message_id",
            "delivered",
            "attempts",
            "created_at",
            "updated_at",
            "before",
            "after",
        )


class OutboundViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows Outbound models to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Outbound.objects.all()
    serializer_class = OutboundSerializer
    filterset_class = OutboundFilter
    filter_backends = (filters.DjangoFilterBackend, OrderingFilter)
    ordering_fields = ("created_at",)
    ordering = ("-created_at",)

    def create(self, *args, **kwargs):
        return super(OutboundViewSet, self).create(*args, **kwargs)


class InboundFilter(filters.FilterSet):
    from_addr = MultipleFilter(field_name="from_addr")
    from_identity = MultipleFilter(field_name="from_identity")

    class Meta:
        model = Inbound
        fields = (
            "message_id",
            "in_reply_to",
            "to_addr",
            "content",
            "transport_name",
            "transport_type",
            "created_at",
            "updated_at",
        )


class InboundPreprocessor(object):
    def pop_from_address(self, data, channel):
        if channel.channel_type == Channel.VUMI_TYPE:
            return data.pop("from_addr", None)
        elif channel.channel_type == Channel.JUNEBUG_TYPE:
            return data.pop("from", None)
        elif channel.channel_type == Channel.WASSUP_API_TYPE:
            return data.get("data", {}).pop("from_addr", None)
        elif channel.channel_type == Channel.WHATSAPP_API_TYPE:
            return data.pop("from", None)

    def is_close_event(self, data, channel):
        if channel.channel_type == Channel.VUMI_TYPE:
            return data.get("session_event") == "close"
        elif channel.channel_type == Channel.JUNEBUG_TYPE:
            return data.get("channel_data", {}).get("session_event") == "close"
        # Wassup/WhatsApp doesn't have sessions
        return False

    def get_related_outbound_id(self, data, channel):
        if channel.channel_type == Channel.VUMI_TYPE:
            return data.get("in_reply_to")
        elif channel.channel_type == Channel.JUNEBUG_TYPE:
            return data.get("reply_to")
        return None

    def get_or_create_identity(self, msisdn):
        result = get_identity_by_address(msisdn)

        if result:
            return result[0]["id"]
        else:
            return create_identity(
                {
                    "details": {
                        "default_addr_type": "msisdn",
                        "addresses": {"msisdn": {msisdn: {"default": True}}},
                    }
                }
            )["id"]

    def preprocess_inbound(self, data, channel):
        msisdn = self.pop_from_address(data, channel)
        if msisdn is None:
            return
        msisdn = e_164(msisdn)
        data["from_identity"] = self.get_or_create_identity(msisdn)

        if channel.concurrency_limit == 0:
            return

        related_outbound = self.get_related_outbound_id(data, channel)
        if self.is_close_event(data, channel) and related_outbound:
            try:
                message = Outbound.objects.get(vumi_message_id=related_outbound)
            except (ObjectDoesNotExist, MultipleObjectsReturned):
                message = (
                    Outbound.objects.filter(to_identity=data["from_identity"])
                    .order_by("-created_at")
                    .last()
                )
            if message:
                ConcurrencyLimiter.decr_message_count(channel, message.last_sent_time)


preprocess_inbound = InboundPreprocessor().preprocess_inbound


class InboundViewSet(viewsets.ModelViewSet):

    """
    API endpoint that allows Inbound models to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Inbound.objects.all()
    filterset_class = InboundFilter
    filter_backends = (filters.DjangoFilterBackend, OrderingFilter)
    ordering_fields = ("created_at",)
    ordering = ("-created_at",)

    def get_serializer_class(self):
        if self.action == "create":
            if self.channel.channel_type == Channel.VUMI_TYPE:
                return InboundSerializer
            elif self.channel.channel_type == Channel.JUNEBUG_TYPE:
                return JunebugInboundSerializer
            elif self.channel.channel_type == Channel.WASSUP_API_TYPE:
                return WassupInboundSerializer
            elif self.channel.channel_type == Channel.WHATSAPP_API_TYPE:
                return WhatsAppInboundSerializer
        return InboundSerializer

    def create(self, request, *args, **kwargs):
        if not kwargs.get("channel_id"):
            self.channel = Channel.objects.get(default=True)
        else:
            self.channel = Channel.objects.get(channel_id=kwargs.get("channel_id"))

        preprocess_inbound(request.data, self.channel)
        return super(InboundViewSet, self).create(request, *args, **kwargs)


def fire_delivery_hook(outbound):
    outbound.refresh_from_db()
    # Only fire if the message has been delivered or we've reached max attempts
    if (
        not outbound.delivered
        and outbound.attempts < settings.MESSAGE_SENDER_MAX_RETRIES
    ):
        return

    payload = {
        "outbound_id": str(outbound.id),
        "delivered": outbound.delivered,
        "to_addr": outbound.to_addr,
    }
    if hasattr(outbound, "to_identity"):
        payload["identity"] = outbound.to_identity

    if payload["to_addr"] is None and payload.get("identity", None) is None:
        raise InvalidMessage(outbound)

    # Becaues the Junebug event endpoint has no authentication, we get an
    # AnonymousUser object for the user. So we have to manually find all of the
    # hook events, ignoring the user, and deliver them.
    hooks = Hook.objects.filter(event="outbound.delivery_report")
    for hook in hooks:
        hook.deliver_hook(None, payload_override={"hook": hook.dict(), "data": payload})


def decr_message_count(message):
    if message.channel:
        channel = message.channel
    else:
        channel = Channel.objects.get(default=True)
    ConcurrencyLimiter.decr_message_count(channel, message.last_sent_time)


outbound_event_total = Counter(
    "outbound_event_total", "Number of Outbound events", ["type", "channel"]
)


def process_event(message_id, event_type, event_detail, timestamp):
    """
    Processes an event of the given details, returning a (success, message) tuple
    """
    # Load message
    try:
        message = Outbound.objects.select_related("channel").get(
            vumi_message_id=message_id
        )
    except ObjectDoesNotExist:
        return (False, "Cannot find message for ID {}".format(message_id))

    if event_type == "ack":
        message.delivered = True
        message.to_addr = ""
        message.metadata["ack_timestamp"] = timestamp
        message.metadata["ack_reason"] = event_detail
        message.save(update_fields=["delivered", "to_addr", "metadata"])
    elif event_type == "nack":
        message.metadata["nack_timestamp"] = timestamp
        message.metadata["nack_reason"] = event_detail
        message.save(update_fields=["metadata"])

        decr_message_count(message)

        send_message.delay(str(message.id))
    elif event_type == "delivery_succeeded":
        message.delivered = True
        message.to_addr = ""
        message.metadata["delivery_timestamp"] = timestamp
        message.metadata["delivery_reason"] = event_detail
        message.save(update_fields=["delivered", "metadata", "to_addr"])
    elif event_type == "delivery_failed":
        message.metadata["delivery_failed_reason"] = event_detail
        message.metadata["delivery_failed_timestamp"] = timestamp
        message.save(update_fields=["metadata"])

        decr_message_count(message)

        send_message.delay(str(message.id))
    elif event_type == "read":
        message.delivered = True
        message.to_addr = ""
        message.metadata["read_timestamp"] = timestamp
        message.save(update_fields=["delivered", "to_addr", "metadata"])

    outbound_event_total.labels(event_type, message.channel_id).inc()
    fire_delivery_hook(message)

    return (True, "Event processed")


class EventListener(APIView):

    """
    Triggers updates to outbound messages based on event data from Vumi
    """

    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        """
        Checks for expect event types before continuing
        """
        serializer = EventSerializer(data=request.data)

        if not serializer.is_valid():
            return Response(
                {"accepted": False, "reason": serializer.errors}, status=400
            )

        data = serializer.validated_data

        event_type = {
            "ack": "ack",
            "nack": "nack",
            "delivery_report": "delivery_succeeded",
        }.get(data["event_type"])

        accepted, reason = process_event(
            data["user_message_id"], event_type, data["nack_reason"], data["timestamp"]
        )

        return Response(
            {"accepted": accepted, "reason": reason}, status=200 if accepted else 400
        )

    # TODO make this work in test harness, works in production
    # def perform_create(self, serializer):
    #     serializer.save(created_by=self.request.user,
    #                     updated_by=self.request.user)
    #
    # def perform_update(self, serializer):
    #     serializer.save(updated_by=self.request.user)


class JunebugEventListener(APIView):

    """
    Triggers updates to outbound messages based on event data from Junebug
    """

    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        """
        Updates the message from the event data.
        """
        serializer = JunebugEventSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                {"accepted": False, "reason": serializer.errors}, status=400
            )
        data = serializer.validated_data

        event_type = {
            "submitted": "ack",
            "rejected": "nack",
            "delivery_succeeded": "delivery_succeeded",
            "delivery_failed": "delivery_failed",
        }.get(data["event_type"])

        accepted, reason = process_event(
            data["message_id"], event_type, data["event_details"], data["timestamp"]
        )

        return Response(
            {"accepted": accepted, "reason": reason}, status=200 if accepted else 400
        )


class WassupEventListener(APIView):
    """
    Triggers updates to outbound messages based on event data from Wassup
    """

    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        serializer = WassupEventSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                {"accepted": False, "reason": serializer.errors}, status=400
            )
        data = serializer.validated_data

        dispatcher = {"message.direct_outbound.status": self.handle_status}
        handler = dispatcher.get(data["hook"]["event"])
        return handler(data["hook"], data["data"])

    def handle_status(self, hook, data):
        event_type = {
            "sent": "ack",
            "unsent": "nack",
            "delivered": "delivery_succeeded",
            "failed": "delivery_failed",
        }.get(data["status"])

        accepted, reason = process_event(
            data["message_uuid"],
            event_type,
            data.get("description"),
            data.get("timestamp"),
        )
        return Response(
            {"accepted": accepted, "reason": reason}, status=200 if accepted else 400
        )


class WhatsAppEventListener(APIView):
    permission_classes = (AllowAny,)

    def validate_signature(self, channel, request):
        secret = channel.configuration.get("HMAC_SECRET")
        if not secret:
            raise AuthenticationFailed(
                "No HMAC_SECRET set on channel {}".format(channel.channel_id)
            )

        signature = request.META.get("HTTP_X_ENGAGE_HOOK_SIGNATURE")
        if not signature:
            raise AuthenticationFailed("X-Engage-Hook-Signature header required")

        h = hmac.new(secret.encode(), request.body, sha256)
        if not hmac.compare_digest(base64.b64encode(h.digest()).decode(), signature):
            raise AuthenticationFailed("Invalid hook signature")

    def handle_event(self, serializer):
        data = serializer.validated_data

        event_type = {
            "sent": "ack",
            "delivered": "delivery_succeeded",
            "failed": "delivery_failed",
            "read": "read",
        }.get(data["status"])

        accepted, reason = process_event(data["id"], event_type, "", data["timestamp"])
        return {"accepted": accepted, "reason": reason, "id": data["id"]}

    def handle_inbound(self, serializer):
        serializer.save()
        return {"accepted": True, "id": serializer.data["id"]}

    def post(self, request, channel_id, *args, **kwargs):
        channel = get_object_or_404(Channel, pk=channel_id)
        self.validate_signature(channel, request)

        serializer = WhatsAppWebhookSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                {"accepted": False, "reason": serializer.errors}, status=400
            )

        events = []
        for event in serializer.validated_data.get("statuses", []):
            event_serializer = WhatsAppEventSerializer(data=event)
            if not event_serializer.is_valid():
                events.append(
                    {
                        "accepted": False,
                        "reason": event_serializer.errors,
                        "id": event.get("id"),
                    }
                )
                continue
            events.append(self.handle_event(event_serializer))

        inbounds = []
        for inbound in serializer.validated_data.get("messages", []):
            preprocess_inbound(inbound, channel)
            inbound_serializer = WhatsAppInboundSerializer(data=inbound)
            if not inbound_serializer.is_valid():
                inbounds.append(
                    {
                        "accepted": False,
                        "reason": inbound_serializer.errors,
                        "id": inbound.get("id"),
                    }
                )
                continue
            inbounds.append(self.handle_inbound(inbound_serializer))

        accepted = all(e["accepted"] for e in events) and all(
            i["accepted"] for i in inbounds
        )
        return Response(
            {"accepted": accepted, "messages": inbounds, "statuses": events},
            status=200 if accepted else 400,
        )


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {"metrics_available": get_available_metrics()}
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        # Uncomment line below if scheduled metrics are added
        # scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_message_sender
        import django
        import rest_framework

        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_message_sender.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__,
                },
            },
        }
        return Response(resp, status=status)


class FailedTaskViewSet(
    mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet
):
    permission_classes = (IsAuthenticated,)
    queryset = OutboundSendFailure.objects.all()
    serializer_class = OutboundSendFailureSerializer
    pagination_class = IdCursorPagination

    def create(self, request):
        status = 201
        resp = {"requeued_failed_tasks": True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)


class AggregateOutboundViewSet(viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = AggregateOutbounds.objects.all()
    serializer_class = AggregateOutboundSerializer

    def create(self, request):
        serializer = self.get_serializer_class()(data=request.data)
        serializer.is_valid(raise_exception=True)
        start = serializer.validated_data.get("start", None)
        end = serializer.validated_data.get("end", None)
        if not end:
            end = datetime.now().date()
        if not start:
            diff = timedelta(days=settings.AGGREGATE_OUTBOUND_BACKTRACK)
            start = (datetime.now() - diff).date()
        aggregate_outbounds.delay(start.isoformat(), end.isoformat())
        return Response({"aggregate_outbounds": True}, status=202)


class ArchivedOutboundViewSet(viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = ArchivedOutbounds.objects.all()
    serializer_class = ArchivedOutboundSerializer

    def create(self, request):
        serializer = self.get_serializer_class()(data=request.data)
        serializer.is_valid(raise_exception=True)
        start = serializer.validated_data["start"]
        end = serializer.validated_data["end"]
        archive_outbound.delay(start.isoformat(), end.isoformat())
        return Response({"archived_outbounds": True}, status=202)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-03-22 07:48
from __future__ import unicode_literals

from django.db import migrations, models
from django.db.backends.postgresql.schema import DatabaseSchemaEditor
import django.db.models.deletion


def modify_fk_constraint(apps, schema_editor):
    """
    Delete's the current foreign key contraint on the outbound field, and adds
    it again, but this time with an ON DELETE clause
    """
    model = apps.get_model("message_sender", "OutboundSendFailure")
    table = model._meta.db_table

    with schema_editor.connection.cursor() as cursor:
        constraints = schema_editor.connection.introspection.get_constraints(
            cursor, table
        )
        [constraint] = filter(lambda c: c[1]["foreign_key"], constraints.items())
        [name, _] = constraint
    sql_delete_fk = (
        "SET CONSTRAINTS {name} IMMEDIATE; "
        "ALTER TABLE {table} DROP CONSTRAINT {name}"
    ).format(table=schema_editor.quote_name(table), name=schema_editor.quote_name(name))
    schema_editor.execute(sql_delete_fk)

    field = model.outbound.field
    to_table = field.remote_field.model._meta.db_table
    to_column = field.remote_field.model._meta.get_field(
        field.remote_field.field_name
    ).column
    sql_create_fk = (
        "ALTER TABLE {table} ADD CONSTRAINT {name} FOREIGN KEY "
        "({column}) REFERENCES {to_table} ({to_column}) "
        "ON DELETE CASCADE {deferrable};"
    ).format(
        table=schema_editor.quote_name(table),
        name=schema_editor.quote_name(name),
        column=schema_editor.quote_name(field.column),
        to_table=schema_editor.quote_name(to_table),
        to_column=schema_editor.quote_name(to_column),
        deferrable=schema_editor.connection.ops.deferrable_sql(),
    )
    schema_editor.execute(sql_create_fk)


class Migration(migrations.Migration):

    dependencies = [
        ("message_sender", "0007_outboundsendfailure"),
        ("message_sender", "0019_auto_20180319_1446"),
    ]

    operations = [
        migrations.AlterField(
            model_name="outboundsendfailure",
            name="outbound",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.DO_NOTHING,
                to="message_sender.Outbound",
            ),
        ),
        migrations.RunPython(modify_fk_constraint),
    ]



# Generated by Django 2.1.2 on 2018-10-30 15:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0023_auto_20180926_0741")]

    operations = [
        migrations.AlterField(
            model_name="inbound",
            name="content",
            field=models.CharField(blank=True, max_length=4096, null=True),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-08 07:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0006_add_created_at_index")]

    operations = [
        migrations.AlterField(
            model_name="outbound",
            name="to_addr",
            field=models.CharField(db_index=True, max_length=500),
        )
    ]



from datetime import datetime
from pytz import UTC
from unittest import TestCase

from seed_message_sender.utils import is_in_time_interval


class IsInTimeIntervalTests(TestCase):
    def test_is_in_time_interval(self):
        """
        Returns whether the timestamp is within the time interval
        """
        self.assertEqual(
            is_in_time_interval(
                "09:00:00+0200/17:00:00Z", datetime(2019, 1, 1, 8, 0, 0, tzinfo=UTC)
            ),
            (True, datetime(2019, 1, 1, 8, 0, 0, tzinfo=UTC)),
        )
        self.assertEqual(
            is_in_time_interval(
                "09:00:00+0200/17:00:00Z", datetime(2019, 1, 1, 18, 0, 0, tzinfo=UTC)
            ),
            (False, datetime(2019, 1, 2, 7, 0, 0, tzinfo=UTC)),
        )
        self.assertEqual(
            is_in_time_interval(
                "09:00:00+0200/17:00:00Z", datetime(2019, 1, 1, 6, 0, 0, tzinfo=UTC)
            ),
            (False, datetime(2019, 1, 1, 7, 0, 0, tzinfo=UTC)),
        )



from django.core.cache import caches
from functools import partial
from rest_framework.authentication import TokenAuthentication


locmem_cache = caches["locmem"]


class CachedTokenAuthentication(TokenAuthentication):
    def authenticate_credentials(self, key):
        """
        Does a cached lookup for a user for the given token
        """
        return locmem_cache.get_or_set(
            "authtoken:{}".format(key), partial(super().authenticate_credentials, key)
        )



# Generated by Django 2.1.2 on 2018-11-16 08:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0024_increase_inbound_message_content_limit")]

    operations = [
        migrations.AlterField(
            model_name="inbound",
            name="content",
            field=models.TextField(blank=True, null=True),
        )
    ]



import os
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path
from message_sender import views
from django_prometheus import exports as django_prometheus
from rest_framework.authtoken.views import obtain_auth_token
from rest_framework.documentation import include_docs_urls
from seed_message_sender.decorators import internal_only

admin.site.site_header = os.environ.get(
    "MESSAGE_SENDER_TITLE", "Seed Message Sender Admin"
)


urlpatterns = [
    path("admin/", admin.site.urls),
    path("api/auth/", include("rest_framework.urls", namespace="rest_framework")),
    path("api/token-auth/", obtain_auth_token),
    path("api/metrics/", views.MetricsView.as_view()),
    path("api/health/", views.HealthcheckView.as_view()),
    path("", include("message_sender.urls")),
    path("docs/", include_docs_urls()),
    path(
        "metrics", internal_only(django_prometheus.ExportToDjangoView), name="metrics"
    ),
]


if settings.DEBUG is True:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from django.conf.urls import url
from django.urls import include, path
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r"inbound", views.InboundViewSet)
router.register(r"outbound", views.OutboundViewSet)
router.register(r"webhook", views.HookViewSet)
router.register(r"failed-tasks", views.FailedTaskViewSet)
router.register(r"aggregate-outbounds", views.AggregateOutboundViewSet)
router.register(r"archive-outbounds", views.ArchivedOutboundViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    path("api/v1/events", views.EventListener.as_view()),
    path(
        "api/v1/events/junebug",
        views.JunebugEventListener.as_view(),
        name="junebug-events",
    ),
    path(
        "api/v1/events/wassup",
        views.WassupEventListener.as_view(),
        name="wassup-events",
    ),
    path(
        "api/v1/events/whatsapp/<channel_id>",
        views.WhatsAppEventListener.as_view(),
        name="whatsapp-events",
    ),
    path("api/v1/user/token/", views.UserView.as_view(), name="create-user-token"),
    url(
        "^api/v1/inbound/(?P<channel_id>\w+)/$",
        views.InboundViewSet.as_view({"post": "create"}),
        name="channels-inbound",
    ),
    path("api/v1/", include(router.urls)),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-05-04 14:46
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("message_sender", "0020_outboundsendfailure_db_backed_fk_constraint")
    ]

    operations = [
        migrations.AlterField(
            model_name="channel",
            name="channel_type",
            field=models.CharField(
                choices=[
                    (b"junebug", b"Junebug"),
                    (b"vumi", b"Vumi"),
                    (b"http_api", b"HTTP API"),
                    (b"wassup", b"Wassup API"),
                ],
                default=b"junebug",
                max_length=20,
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 16:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0012_channel_insert")]

    operations = [
        migrations.AddField(
            model_name="outbound",
            name="channel",
            field=models.ForeignKey(
                null=True,
                on_delete=models.deletion.CASCADE,
                to="message_sender.Channel",
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:57
from __future__ import unicode_literals
import os

from django.db import migrations
from django.conf import settings


def insert_channels(apps, schema_editor):

    default = True

    Channel = apps.get_model("message_sender.channel")

    if settings.MESSAGE_BACKEND_TEXT == "vumi":
        if not Channel.objects.filter(channel_id="VUMI_TEXT").exists():
            vumi_url = os.environ.get("MESSAGE_SENDER_VUMI_API_URL_TEXT")
            vumi_account_key = os.environ.get("MESSAGE_SENDER_VUMI_ACCOUNT_KEY_TEXT")
            vumi_conv_key = os.environ.get("MESSAGE_SENDER_VUMI_CONVERSATION_KEY_TEXT")
            vumi_acc_token = os.environ.get("MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_TEXT")

            if vumi_url and vumi_account_key and vumi_conv_key and vumi_acc_token:
                channel = Channel()
                channel.channel_id = "VUMI_TEXT"
                channel.default = default
                channel.channel_type = "vumi"
                channel.concurrency_limit = settings.CONCURRENT_TEXT_LIMIT
                channel.message_delay = settings.TEXT_MESSAGE_DELAY
                channel.message_timeout = settings.TEXT_MESSAGE_TIMEOUT
                channel.configuration = {
                    "VUMI_API_URL": vumi_url,
                    "VUMI_ACCOUNT_KEY": vumi_account_key,
                    "VUMI_CONVERSATION_KEY": vumi_conv_key,
                    "VUMI_ACCOUNT_TOKEN": vumi_acc_token,
                }

                channel.save()

                default = False

    elif settings.MESSAGE_BACKEND_TEXT == "junebug":
        if not Channel.objects.filter(channel_id="JUNEBUG_TEXT").exists():
            june_url = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_URL_TEXT")
            june_auth = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_AUTH_TEXT")
            june_from = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_FROM_TEXT")

            if june_url and june_auth and june_from:
                channel = Channel()
                channel.channel_id = "JUNEBUG_TEXT"
                channel.default = default
                channel.channel_type = "junebug"
                channel.concurrency_limit = settings.CONCURRENT_TEXT_LIMIT
                channel.message_delay = settings.TEXT_MESSAGE_DELAY
                channel.message_timeout = settings.TEXT_MESSAGE_TIMEOUT
                channel.configuration = {
                    "JUNEBUG_API_URL": june_url,
                    "JUNEBUG_API_AUTH": june_auth,
                    "JUNEBUG_API_FROM": june_from,
                }
                channel.save()

                default = False

    if settings.MESSAGE_BACKEND_VOICE == "vumi":
        if not Channel.objects.filter(channel_id="VUMI_VOICE").exists():
            vumi_url = os.environ.get("MESSAGE_SENDER_VUMI_API_URL_VOICE")
            vumi_account_key = os.environ.get("MESSAGE_SENDER_VUMI_ACCOUNT_KEY_VOICE")
            vumi_conv_key = os.environ.get("MESSAGE_SENDER_VUMI_CONVERSATION_KEY_VOICE")
            vumi_acc_token = os.environ.get("MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_VOICE")

            if vumi_url and vumi_account_key and vumi_conv_key and vumi_acc_token:
                channel = Channel()
                channel.channel_id = "VUMI_VOICE"
                channel.default = default
                channel.channel_type = "vumi"
                channel.concurrency_limit = settings.CONCURRENT_VOICE_LIMIT
                channel.message_delay = settings.VOICE_MESSAGE_DELAY
                channel.message_timeout = settings.VOICE_MESSAGE_TIMEOUT
                channel.configuration = {
                    "VUMI_API_URL": vumi_url,
                    "VUMI_ACCOUNT_KEY": vumi_account_key,
                    "VUMI_CONVERSATION_KEY": vumi_conv_key,
                    "VUMI_ACCOUNT_TOKEN": vumi_acc_token,
                }
                channel.save()

                default = False

    elif settings.MESSAGE_BACKEND_VOICE == "junebug":
        if not Channel.objects.filter(channel_id="JUNEBUG_VOICE").exists():
            june_url = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_URL_VOICE")
            june_auth = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_AUTH_VOICE")
            june_from = os.environ.get("MESSAGE_SENDER_JUNEBUG_API_FROM_VOICE")

            if june_url and june_auth and june_from:
                channel = Channel()
                channel.channel_id = "JUNEBUG_VOICE"
                channel.default = default
                channel.channel_type = "junebug"
                channel.concurrency_limit = settings.CONCURRENT_VOICE_LIMIT
                channel.message_delay = settings.VOICE_MESSAGE_DELAY
                channel.message_timeout = settings.VOICE_MESSAGE_TIMEOUT
                channel.configuration = {
                    "JUNEBUG_API_URL": june_url,
                    "JUNEBUG_API_AUTH": june_auth,
                    "JUNEBUG_API_FROM": june_from,
                }
                channel.save()


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0011_channel")]

    operations = [migrations.RunPython(insert_channels)]



import json
import os
import re
from copy import deepcopy

import pkg_resources
import requests
from django.conf import settings
from django.urls import reverse
from go_http.send import HttpApiSender
from rest_hooks.models import Hook
from requests import exceptions as requests_exceptions
from six.moves import urllib_parse

from .models import Channel
from .utils import make_absolute_url


class FactoryException(Exception):
    pass


class HttpApiSenderException(Exception):
    pass


# HttpApiSender doesn't take message metadata,
# but we need the metadata for the WhatsApp channel, so we remove it here
class BaseHttpApiSender(HttpApiSender):
    def send_text(self, *args, **kwargs):
        kwargs.pop("metadata", None)
        return super().send_text(*args, **kwargs)


class VumiHttpApiSender(BaseHttpApiSender):
    def send_image(self, to_addr, content, image_url=None):
        raise HttpApiSenderException("Sending images not available on this channel.")


class GenericHttpApiSender(VumiHttpApiSender):
    def __init__(
        self,
        url,
        auth=None,
        from_addr=None,
        session=None,
        override_payload=None,
        strip_filepath=False,
    ):
        """
        :param url str: The URL for the HTTP API channel
        :param auth tuple: (username, password) or anything
            accepted by the requests library. Defaults to None.
        :param session requests.Session: A requests session. Defaults to None
        :param from_addr str: The from address for all messages. Defaults to
            None
        :param override_payload dict: This is the format of the payload that
            needs to be sent to the URL. It willl be populated from the
            original payload. Defaults to None
        :param strip_filepath boolean: This should be true if we only need to
            send the filename to the api.
        """
        self.api_url = url
        self.auth = tuple(auth) if isinstance(auth, list) else auth
        self.from_addr = from_addr
        if session is None:
            session = requests.Session()
        self.session = session
        self.override_payload = override_payload
        self.strip_filepath = strip_filepath

    def _get_filename(self, path):
        """
        This function gets the base filename from the path, if a language code
        is present the filename will start from there.
        """
        match = re.search("[a-z]{2,3}_[A-Z]{2}", path)

        if match:
            start = match.start(0)
            filename = path[start:]
        else:
            filename = os.path.basename(path)

        return filename

    def _raw_send(self, py_data):
        headers = {"content-type": "application/json; charset=utf-8"}

        channel_data = py_data.get("helper_metadata", {})
        channel_data["session_event"] = py_data.get("session_event")

        url = channel_data.get("voice", {}).get("speech_url")
        if self.strip_filepath and url:
            if not isinstance(url, (list, tuple)):
                channel_data["voice"]["speech_url"] = self._get_filename(url)
            else:
                channel_data["voice"]["speech_url"] = []
                for item in url:
                    channel_data["voice"]["speech_url"].append(self._get_filename(item))

        data = {
            "to": py_data["to_addr"],
            "from": self.from_addr,
            "content": py_data["content"],
            "channel_data": channel_data,
        }

        data = self._override_payload(data)

        data = json.dumps(data)
        r = self.session.post(
            self.api_url,
            auth=self.auth,
            data=data,
            headers=headers,
            timeout=settings.DEFAULT_REQUEST_TIMEOUT,
        )
        r.raise_for_status()
        res = r.json()
        return res.get("result", {})

    def _override_payload(self, payload):
        """
        This function transforms the payload into a new format using the
        self.override_payload property.
        """
        if self.override_payload:
            old_payload = payload

            def get_value(data, key):
                try:
                    parent_key, nested_key = key.split(".", 1)
                    return get_value(data.get(parent_key, {}), nested_key)
                except ValueError:
                    return data.get(key, key)

            def set_values(data):
                for key, value in data.items():
                    if isinstance(value, dict):
                        set_values(value)
                    else:
                        data[key] = get_value(old_payload, value)

            payload = deepcopy(self.override_payload)
            set_values(payload)

        return payload

    def fire_metric(self, metric, value, agg="last"):
        raise HttpApiSenderException("Metrics sending not supported")


class JunebugApiSender(GenericHttpApiSender):
    def _raw_send(self, py_data):
        headers = {"content-type": "application/json; charset=utf-8"}

        channel_data = py_data.get("helper_metadata", {})
        channel_data["session_event"] = py_data.get("session_event")

        data = {
            "to": py_data["to_addr"],
            "from": self.from_addr,
            "content": py_data["content"],
            "channel_data": channel_data,
            "event_url": make_absolute_url(reverse("junebug-events")),
        }

        data = json.dumps(data)
        r = self.session.post(
            self.api_url,
            auth=self.auth,
            data=data,
            headers=headers,
            timeout=settings.DEFAULT_REQUEST_TIMEOUT,
        )
        r.raise_for_status()
        res = r.json()
        return res.get("result", {})


class WassupApiSenderException(Exception):
    pass


WASSUP_SESSIONS = {}


class WassupApiSender(object):
    def __init__(
        self, api_url, token, hsm_uuid, hsm_disabled, number=None, session=None
    ):
        self.api_url = api_url
        self.token = token
        self.hsm_uuid = hsm_uuid
        self.hsm_disabled = hsm_disabled
        self.number = number

        distribution = pkg_resources.get_distribution("seed_message_sender")

        # reuse sessions on tokens to make use of SSL keep-alive
        # but keep some separation around auth
        self.session = session or WASSUP_SESSIONS.setdefault(token, requests.Session())
        self.session.headers.update(
            {
                "Authorization": "Token %s" % (self.token,),
                "User-Agent": "SeedMessageSender/%s" % (distribution.version,),
            }
        )

    def send_text(self, to_addr, content, session_event=None, metadata=None):
        if self.hsm_disabled:
            if not self.number:
                raise WassupApiSenderException(
                    "Cannot send a non hsm message if a number is not " "specified."
                )

            response = self.session.post(
                urllib_parse.urljoin(self.api_url, "/api/v1/messages/"),
                json={"number": self.number, "content": content, "to_addr": to_addr},
            )
        else:
            response = self.session.post(
                urllib_parse.urljoin(
                    self.api_url, "/api/v1/hsms/%s/send/" % (self.hsm_uuid,)
                ),
                json={"to_addr": to_addr, "localizable_params": [{"default": content}]},
            )

        response.raise_for_status()
        data = response.json()
        # the SendMessage task expects the sender to return a dict with
        # a ``message_id`` field set. I'm injecting that here manually to
        # comply
        data.update({"message_id": data["uuid"]})
        return data

    def send_image(self, to_addr, content, image_url=None):
        if not self.number:
            raise WassupApiSenderException(
                "Cannot send a image file if a number is not specified."
            )

        image_file = requests.get(image_url, stream=True)
        image_file.raise_for_status()

        content_type = image_file.headers["content-type"]
        image_stream = image_file.raw
        image_name = image_url.split("/")[-1]

        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/api/v1/messages/"),
            files={"image_attachment": (image_name, image_stream, content_type, {})},
            data={
                "to_addr": to_addr,
                "number": self.number,
                "image_attachment_caption": content,
            },
        )
        response.raise_for_status()
        data = response.json()
        data.update({"message_id": data["uuid"]})
        return data

    def send_voice(
        self, to_addr, content, speech_url=None, wait_for=None, session_event=None
    ):
        if not self.number:
            raise WassupApiSenderException(
                "Cannot send an audio file if a number is not specified."
            )

        audio_file = requests.get(speech_url, stream=True)
        audio_file.raise_for_status()

        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/api/v1/messages/"),
            files={"audio_attachment": audio_file.raw},
            data={"to_addr": to_addr, "number": self.number},
        )
        response.raise_for_status()
        data = response.json()
        data.update({"message_id": data["uuid"]})
        return data

    def fire_metric(self, metric, value, agg="last"):
        raise WassupApiSenderException("Metrics sending not supported")


class WhatsAppApiSenderException(Exception):
    pass


WHATSAPP_SESSIONS = {}


class WhatsAppApiSender(object):
    def __init__(
        self, api_url, token, hsm_namespace, hsm_element_name, ttl, session=None
    ):
        self.api_url = api_url
        self.token = token
        self.hsm_namespace = hsm_namespace
        self.hsm_element_name = hsm_element_name
        self.ttl = ttl

        distribution = pkg_resources.get_distribution("seed_message_sender")

        # reuse sessions on tokens to make use of SSL keep-alive
        # but keep some separation around auth
        self.session = session or WHATSAPP_SESSIONS.setdefault(
            token, requests.Session()
        )
        self.session.headers.update(
            {
                "Authorization": "Bearer %s" % (self.token,),
                "User-Agent": "SeedMessageSender/%s" % (distribution.version,),
            }
        )

    def fire_failed_contact_lookup(self, msisdn):
        """
        Fires a webhook in the event of a failed WhatsApp contact lookup.
        """
        payload = {"address": msisdn}
        # We cannot user the raw_hook_event here, because we don't have a user, so we
        # manually filter and send the hooks for all users
        hooks = Hook.objects.filter(event="whatsapp.failed_contact_check")
        for hook in hooks:
            hook.deliver_hook(
                None, payload_override={"hook": hook.dict(), "data": payload}
            )

    def get_contact(self, msisdn):
        """
        Returns the WhatsApp ID for the given MSISDN
        """
        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/v1/contacts"),
            json={"blocking": "wait", "contacts": [msisdn]},
        )
        response.raise_for_status()
        whatsapp_id = response.json()["contacts"][0].get("wa_id")
        if not whatsapp_id:
            self.fire_failed_contact_lookup(msisdn)
        return whatsapp_id

    def send_hsm(self, whatsapp_id, content):
        data = {
            "to": whatsapp_id,
            "type": "hsm",
            "hsm": {
                "namespace": self.hsm_namespace,
                "element_name": self.hsm_element_name,
                "localizable_params": [{"default": content}],
            },
        }

        if self.ttl is not None:
            data["ttl"] = self.ttl
        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/v1/messages"), json=data
        )
        return self.return_response(response)

    def send_custom_hsm(self, whatsapp_id, template_name, language, variables):
        """
        Sends an HSM with more customizable fields than the send_hsm function
        """
        data = {
            "to": whatsapp_id,
            "type": "hsm",
            "hsm": {
                "namespace": self.hsm_namespace,
                "element_name": template_name,
                "language": {"policy": "deterministic", "code": language},
                "localizable_params": [{"default": variable} for variable in variables],
            },
        }

        if self.ttl is not None:
            data["ttl"] = self.ttl
        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/v1/messages"), json=data
        )
        return self.return_response(response)

    def send_text_message(self, whatsapp_id, content):
        response = self.session.post(
            urllib_parse.urljoin(self.api_url, "/v1/messages"),
            json={"to": whatsapp_id, "text": {"body": content}},
        )
        return self.return_response(response)

    def return_response(self, response):
        try:
            response.raise_for_status()
        except requests_exceptions.HTTPError as exc:
            resp = exc.response.text

            if not ("1006" in resp and "unknown contact" in resp):
                raise exc

        return response.json()

    def send_text(self, to_addr, content, session_event=None, metadata=None):
        whatsapp_id = to_addr.replace("+", "")

        def send_message():
            if metadata and "template" in metadata:
                template = metadata["template"]
                d = self.send_custom_hsm(
                    whatsapp_id,
                    template["name"],
                    template["language"],
                    template["variables"],
                )
            elif self.hsm_namespace and self.hsm_element_name:
                d = self.send_hsm(whatsapp_id, content)
            else:
                d = self.send_text_message(whatsapp_id, content)
            return d

        data = send_message()

        if (
            "errors" in data
            and data["errors"][0]["code"] == 1006
            and data["errors"][0]["details"] == "unknown contact"
        ):
            whatsapp_id = self.get_contact(to_addr)
            if not whatsapp_id:
                return {"message_id": None}

            data = send_message()

        return {"message_id": data["messages"][0]["id"]}

    def send_image(self, to_addr, content, image_url=None):
        raise WhatsAppApiSenderException("Image sending not supported")

    def send_voice(
        self, to_addr, content, speech_url=None, wait_for=None, session_event=None
    ):
        raise WhatsAppApiSenderException("Voice sending not supported")

    def fire_metric(self, metric, value, agg="last"):
        raise WhatsAppApiSenderException("Metrics sending not supported")


class MessageClientFactory(object):
    @classmethod
    def create(cls, channel=None):
        try:
            if not channel:
                channel = Channel.objects.get(default=True)
        except Channel.DoesNotExist:
            raise FactoryException("Unknown backend type: %r" % (channel,))

        backend_type = channel.channel_type
        handler = getattr(cls, "create_%s_client" % (backend_type,), None)
        if not handler:
            raise FactoryException("Unknown backend type: %r" % (backend_type,))

        return handler(channel)

    @classmethod
    def create_junebug_client(cls, channel):
        return JunebugApiSender(
            channel.configuration.get("JUNEBUG_API_URL"),
            channel.configuration.get("JUNEBUG_API_AUTH"),
            channel.configuration.get("JUNEBUG_API_FROM"),
        )

    @classmethod
    def create_wassup_client(cls, channel):
        return WassupApiSender(
            channel.configuration.get("WASSUP_API_URL"),
            channel.configuration.get("WASSUP_API_TOKEN"),
            channel.configuration.get("WASSUP_API_HSM_UUID"),
            channel.configuration.get("WASSUP_API_HSM_DISABLED", False),
            number=channel.configuration.get("WASSUP_API_NUMBER"),
        )

    @classmethod
    def create_vumi_client(cls, channel):
        return VumiHttpApiSender(
            channel.configuration.get("VUMI_ACCOUNT_KEY"),
            channel.configuration.get("VUMI_CONVERSATION_KEY"),
            channel.configuration.get("VUMI_ACCOUNT_TOKEN"),
            api_url=channel.configuration.get("VUMI_API_URL"),
        )

    @classmethod
    def create_http_api_client(cls, channel):
        return GenericHttpApiSender(
            channel.configuration.get("HTTP_API_URL"),
            channel.configuration.get("HTTP_API_AUTH"),
            channel.configuration.get("HTTP_API_FROM"),
            override_payload=channel.configuration.get("OVERRIDE_PAYLOAD"),
            strip_filepath=channel.configuration.get("STRIP_FILEPATH"),
        )

    @classmethod
    def create_whatsapp_client(cls, channel):
        return WhatsAppApiSender(
            channel.configuration["API_URL"],
            channel.configuration["API_TOKEN"],
            channel.configuration.get("HSM_NAMESPACE"),
            channel.configuration.get("HSM_ELEMENT_NAME"),
            channel.configuration.get("TTL"),
        )



from datetime import datetime
from django.test import TestCase, override_settings

from .utils import daterange, make_absolute_url


class TestMakeAbsoluteUrl(TestCase):
    @override_settings(USE_SSL=True)
    def test_make_absolute_url_ssl_true(self):
        """
        If USE_SSL is True, then then the url should be using https
        """
        self.assertEqual(make_absolute_url("foo"), "https://example.com/foo")
        self.assertEqual(make_absolute_url("/foo"), "https://example.com/foo")

    @override_settings(USE_SSL=False)
    def test_make_absolute_url_ssl_false(self):
        """
        If USE_SSL is False, then then the url should be using http
        """
        self.assertEqual(make_absolute_url("foo"), "http://example.com/foo")
        self.assertEqual(make_absolute_url("/foo"), "http://example.com/foo")


class DateRangeTests(TestCase):
    def test_date_range(self):
        """
        daterange should return the list of dates that falls within the range
        """
        self.assertEqual(
            list(daterange(datetime(2017, 1, 1), datetime(2017, 1, 3))),
            [datetime(2017, 1, 1), datetime(2017, 1, 2), datetime(2017, 1, 3)],
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 11:57
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0010_merge")]

    operations = [
        migrations.CreateModel(
            name="Channel",
            fields=[
                (
                    "channel_id",
                    models.CharField(
                        editable=True, max_length=64, primary_key=True, serialize=False
                    ),
                ),
                (
                    "channel_type",
                    models.CharField(
                        choices=[(b"junebug", b"Junebug"), (b"vumi", b"Vumi")],
                        default=b"junebug",
                        max_length=20,
                    ),
                ),
                ("concurrency_limit", models.IntegerField(default=0)),
                ("message_delay", models.IntegerField(default=0)),
                ("message_timeout", models.IntegerField(default=0)),
                ("default", models.BooleanField(default=False)),
                ("configuration", django.contrib.postgres.fields.jsonb.JSONField()),
            ],
        )
    ]



from __future__ import absolute_import

import os

from celery import Celery

from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "seed_message_sender.settings")

app = Celery("seed_message_sender")

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


@app.task(bind=True)
def debug_task(self):
    print("Request: {0!r}".format(self.request))



from rest_framework import serializers
from rest_hooks.models import Hook

from .models import Channel, Inbound, Outbound, OutboundSendFailure


class OneFieldRequiredValidator:
    def __init__(self, fields):
        self.fields = fields

    def set_context(self, serializer):
        self.is_create = getattr(serializer, "instance", None) is None

    def __call__(self, data):
        if self.is_create:

            for field in self.fields:
                if data.get(field):
                    return

            raise serializers.ValidationError(
                "One of these fields must be populated: %s" % (", ".join(self.fields))
            )


class OutboundSerializer(serializers.HyperlinkedModelSerializer):

    channel = serializers.PrimaryKeyRelatedField(
        queryset=Channel.objects.all(), required=False
    )

    class Meta:
        model = Outbound
        fields = (
            "url",
            "id",
            "version",
            "to_addr",
            "vumi_message_id",
            "content",
            "delivered",
            "attempts",
            "metadata",
            "created_at",
            "updated_at",
            "channel",
            "to_identity",
            "resend",
        )
        validators = [OneFieldRequiredValidator(["to_addr", "to_identity"])]


class OutboundArchiveSerializer(serializers.ModelSerializer):
    class Meta:
        model = Outbound
        fields = (
            "id",
            "to_addr",
            "to_identity",
            "version",
            "content",
            "vumi_message_id",
            "delivered",
            "resend",
            "call_answered",
            "attempts",
            "metadata",
            "channel",
            "updated_at",
            "created_at",
            "last_sent_time",
            "created_by",
            "updated_by",
        )


class InboundSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Inbound
        fields = (
            "url",
            "id",
            "message_id",
            "in_reply_to",
            "to_addr",
            "from_addr",
            "content",
            "transport_name",
            "transport_type",
            "helper_metadata",
            "created_at",
            "updated_at",
            "from_identity",
        )
        validators = [OneFieldRequiredValidator(["from_addr", "from_identity"])]

    def to_internal_value(self, data):
        """
        Adds extra data to the helper_metadata field.
        """
        if "session_event" in data:
            data["helper_metadata"]["session_event"] = data["session_event"]

        return super(InboundSerializer, self).to_internal_value(data)


class JunebugInboundSerializer(serializers.HyperlinkedModelSerializer):
    """
    Maps fields from Junebug onto fields expected by the Inbound model.
    """

    reply_to = serializers.CharField(source="in_reply_to")
    to = serializers.CharField(source="to_addr")
    channel_id = serializers.CharField(source="transport_name")
    channel_data = serializers.JSONField(source="helper_metadata")

    class Meta:
        model = Inbound
        fields = (
            "url",
            "id",
            "message_id",
            "reply_to",
            "to",
            "from_addr",
            "content",
            "channel_id",
            "channel_data",
            "created_at",
            "updated_at",
            "from_identity",
        )
        validators = [OneFieldRequiredValidator(["from", "from_identity"])]


class WassupHookSerializer(serializers.Serializer):
    event = serializers.CharField()


class WassupDataSerializer(serializers.Serializer):
    uuid = serializers.CharField()
    to_addr = serializers.CharField()
    in_reply_to = serializers.CharField(allow_null=True)
    content = serializers.CharField()
    metadata = serializers.JSONField()

    class Meta:
        fields = (
            "uuid",
            "from_addr",
            "from_identity",
            "to_addr",
            "in_reply_to",
            "content",
            "metadata",
        )


class WassupInboundSerializer(serializers.Serializer):
    """
    Maps fields from Junebug onto fields expected by the Inbound model.
    """

    hook = WassupHookSerializer()
    data = WassupDataSerializer()
    from_identity = serializers.CharField()

    class Meta:
        fields = ("hook", "data")

    def create(self, validated_data):
        data = validated_data["data"]
        Inbound.objects.create(
            message_id=data["uuid"],
            in_reply_to=data["in_reply_to"],
            to_addr=data["to_addr"],
            from_identity=validated_data["from_identity"],
            content=data["content"],
            helper_metadata=data["metadata"],
        )
        return validated_data


class WhatsAppInboundSerializer(serializers.Serializer):
    """
    Maps fields from WhatsApp onto the fields expected by the Inbound model.
    """

    id = serializers.CharField()

    class TextSerializer(serializers.Serializer):
        body = serializers.CharField()

    text = TextSerializer()

    # This gets added by the view
    from_identity = serializers.UUIDField(required=False)

    def create(self, validated_data):
        Inbound.objects.create(
            message_id=validated_data["id"],
            from_identity=validated_data["from_identity"],
            content=validated_data["text"]["body"],
            helper_metadata={},
        )
        return validated_data


# Since from is a reserved keyword, we need to set it here
setattr(WhatsAppInboundSerializer, "from", serializers.CharField())


class HookSerializer(serializers.ModelSerializer):
    class Meta:
        model = Hook
        read_only_fields = ("user",)
        fields = "__all__"


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class OutboundSendFailureSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = OutboundSendFailure
        fields = ("url", "id", "outbound", "task_id", "initiated_at", "reason")


class AggregateOutboundSerializer(serializers.Serializer):
    start = serializers.DateField(required=False)
    end = serializers.DateField(required=False)


class ArchivedOutboundSerializer(serializers.Serializer):
    start = serializers.DateField()
    end = serializers.DateField()


class EventSerializer(serializers.Serializer):
    message_type = serializers.ChoiceField(choices=["event"])
    event_type = serializers.ChoiceField(choices=["ack", "nack", "delivery_report"])
    user_message_id = serializers.CharField()
    timestamp = serializers.CharField()
    nack_reason = serializers.JSONField(default="")


class JunebugEventSerializer(serializers.Serializer):
    event_type = serializers.ChoiceField(
        choices=["submitted", "rejected", "delivery_succeeded", "delivery_failed"]
    )
    message_id = serializers.CharField()
    timestamp = serializers.CharField()
    event_details = serializers.JSONField(default="")


class WassupEventSerializer(serializers.Serializer):
    class HookSerializer(serializers.Serializer):
        event = serializers.ChoiceField(choices=["message.direct_outbound.status"])

    hook = HookSerializer()

    class DataSerializer(serializers.Serializer):
        status = serializers.ChoiceField(
            choices=["sent", "unsent", "delivered", "failed"]
        )
        message_uuid = serializers.CharField()
        description = serializers.CharField(default="")
        timestamp = serializers.CharField()

    data = DataSerializer()


class WhatsAppEventSerializer(serializers.Serializer):
    id = serializers.CharField()
    status = serializers.ChoiceField(choices=["sent", "delivered", "read", "failed"])
    timestamp = serializers.CharField()


class WhatsAppWebhookSerializer(serializers.Serializer):
    statuses = serializers.ListField(required=False)
    messages = serializers.ListField(required=False)



import gzip
import json
import os
import pytz
import random
import requests
import time

from celery.exceptions import MaxRetriesExceededError, SoftTimeLimitExceeded
from celery.task import Task
from celery.utils.log import get_task_logger
from rest_hooks.models import Hook

from datetime import datetime, timedelta
from demands import HTTPServiceError
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.core.files import File
from django.db.models import Count, Sum
from django.db.models.signals import post_delete
from django.utils import timezone

from seed_services_client.metrics import MetricsApiClient
from requests import exceptions as requests_exceptions
from requests.exceptions import ConnectionError, HTTPError
from rest_framework.renderers import JSONRenderer
from rest_hooks.models import model_deleted
from seed_message_sender.celery import app
from seed_message_sender.utils import is_in_time_interval

from .factory import MessageClientFactory


from .models import (
    Outbound,
    OutboundSendFailure,
    Channel,
    AggregateOutbounds,
    ArchivedOutbounds,
)
from .serializers import OutboundArchiveSerializer
from seed_message_sender.utils import (
    load_callable,
    get_identity_address,
    get_identity_by_address,
    create_identity,
)
from message_sender.utils import daterange

logger = get_task_logger(__name__)

voice_to_addr_formatter = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
text_to_addr_formatter = load_callable(settings.TEXT_TO_ADDR_FORMATTER)


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay


@app.task(
    autoretry_for=(HTTPError, ConnectionError, SoftTimeLimitExceeded),
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
)
def deliver_hook(target, payload, instance_id=None, hook_id=None, **kwargs):
    """
        target:     the url to receive the payload.
        payload:    a python primitive data structure
        instance_id:   a possibly None "trigger" instance ID
        hook_id:       the ID of defining Hook object
        """
    r = requests.post(
        url=target,
        data=json.dumps(payload),
        headers={
            "Content-Type": "application/json",
            "Authorization": "Token %s" % settings.HOOK_AUTH_TOKEN,
        },
    )
    r.raise_for_status()
    return r.text


def deliver_hook_wrapper(target, payload, instance, hook):
    if instance is not None:
        instance_id = instance.id
    else:
        instance_id = None
    kwargs = dict(
        target=target, payload=payload, instance_id=instance_id, hook_id=hook.id
    )
    deliver_hook.apply_async(kwargs=kwargs)


metric_client = MetricsApiClient(url=settings.METRICS_URL, auth=settings.METRICS_AUTH)


@app.task(
    autoretry_for=(HTTPError, ConnectionError, HTTPServiceError, SoftTimeLimitExceeded),
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
)
def fire_metric(metric_name, metric_value):
    """ Fires a metric using the MetricsApiClient
    """
    metric_value = float(metric_value)
    metric = {metric_name: metric_value}
    metric_client.fire_metrics(**metric)
    return "Fired metric <{}> with value <{}>".format(metric_name, metric_value)


class ConcurrencyLimiter(object):
    BUCKET_SIZE = 60

    @classmethod
    def get_key(cls, channel_id, bucket):
        return "%s_messages_at_%s" % (channel_id, bucket)

    @classmethod
    def get_current_message_count(cls, channel):
        # Sum the values in all the buckets to get the total
        total = 0
        number_of_buckets = channel.message_timeout // cls.BUCKET_SIZE + 1
        bucket = int(time.time() // cls.BUCKET_SIZE)
        for i in range(bucket, bucket - number_of_buckets, -1):
            value = cache.get(cls.get_key(channel.channel_id, i))
            if value:
                total += int(value)
        return total

    @classmethod
    def incr_message_count(cls, channel_id, timeout):
        bucket = int(time.time() // cls.BUCKET_SIZE)
        key = cls.get_key(channel_id, bucket)

        # Add the bucket size to the expiry time so messages that start at
        # the end of the bucket still complete
        if not cache.add(key, 1, timeout + cls.BUCKET_SIZE):
            cache.incr(key)

    @classmethod
    def decr_message_count(cls, channel, msg_time):

        if channel.concurrency_limit == 0:
            return
        timeout = channel.message_timeout

        if not msg_time:
            return

        # Convert from datetime to seconds since epoch
        msg_time = msg_time.replace(tzinfo=None) - msg_time.utcoffset()
        msg_time = (msg_time - datetime(1970, 1, 1)).total_seconds()

        time_since = time.time() - msg_time
        if time_since > timeout:
            return
        bucket = int(msg_time // cls.BUCKET_SIZE)

        key = cls.get_key(channel.channel_id, bucket)
        # Set the expiry time to the timeout minus the time passed since
        # the message was sent.
        if int(cache.get_or_set(key, lambda: 0, timeout - time_since)) > 0:
            cache.decr(key)

    @classmethod
    def manage_limit(cls, task, channel):
        limit = channel.concurrency_limit
        timeout = channel.message_timeout
        delay = channel.message_delay

        if limit > 0:
            if cls.get_current_message_count(channel) >= limit:
                task.retry(countdown=delay)
            cls.incr_message_count(channel.channel_id, timeout)


class SendMessage(Task):

    """
    Task to load and contruct message and send them off
    """

    name = "message_sender.tasks.send_message"
    default_retry_delay = 5
    max_retries = None
    max_error_retries = 5

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def get_client(self, channel=None):
        return MessageClientFactory.create(channel)

    def fire_failed_msisdn_lookup(self, to_identity):
        """
        Fires a webhook in the event of a None to_addr.
        """
        payload = {"to_identity": to_identity}
        hooks = Hook.objects.filter(event="identity.no_address")
        for hook in hooks:
            hook.deliver_hook(
                None, payload_override={"hook": hook.dict(), "data": payload}
            )

    def run(self, message_id, **kwargs):
        """
        Load and contruct message and send them off
        """
        log = self.get_logger(**kwargs)

        if settings.SAFE_TIME_INTERVAL:
            safe, next_time = is_in_time_interval(settings.SAFE_TIME_INTERVAL)
            if not safe:
                self.retry(eta=next_time)

        error_retry_count = kwargs.get("error_retry_count", 0)
        if error_retry_count >= self.max_error_retries:
            raise MaxRetriesExceededError(
                "Can't retry {0}[{1}] args:{2} kwargs:{3}".format(
                    self.name, self.request.id, self.request.args, kwargs
                )
            )

        log.info("Loading Outbound Message <%s>" % message_id)
        try:
            message = Outbound.objects.select_related("channel").get(id=message_id)
        except ObjectDoesNotExist:
            logger.error("Missing Outbound message", exc_info=True)
            return

        if message.attempts < settings.MESSAGE_SENDER_MAX_RETRIES:
            if error_retry_count > 0:
                retry_delay = calculate_retry_delay(error_retry_count)
            else:
                retry_delay = self.default_retry_delay
            log.info("Attempts: %s" % message.attempts)
            # send or resend
            try:
                if not message.channel:
                    channel = Channel.objects.get(default=True)
                else:
                    channel = message.channel

                sender = self.get_client(channel)
                ConcurrencyLimiter.manage_limit(self, channel)

                if not message.to_addr and message.to_identity:
                    message.to_addr = get_identity_address(
                        message.to_identity, use_communicate_through=True
                    )
                    if not message.to_addr:
                        self.fire_failed_msisdn_lookup(message.to_identity)
                        return

                if message.to_addr and not message.to_identity:
                    result = get_identity_by_address(message.to_addr)

                    if result:
                        message.to_identity = result[0]["id"]
                    else:
                        identity = {
                            "details": {
                                "default_addr_type": "msisdn",
                                "addresses": {
                                    "msisdn": {message.to_addr: {"default": True}}
                                },
                            }
                        }
                        identity = create_identity(identity)
                        message.to_identity = identity["id"]

                if "voice_speech_url" in message.metadata:
                    # OBD number of tries metric
                    fire_metric.apply_async(
                        kwargs={
                            "metric_name": "vumimessage.obd.tries.sum",
                            "metric_value": 1.0,
                        }
                    )

                    # Voice message
                    speech_url = message.metadata["voice_speech_url"]
                    vumiresponse = sender.send_voice(
                        voice_to_addr_formatter(message.to_addr),
                        message.content,
                        speech_url=speech_url,
                        session_event="new",
                    )
                    log.info("Sent voice message to <%s>" % message.to_addr)

                elif "image_url" in message.metadata:
                    # Image message
                    image_url = message.metadata["image_url"]
                    vumiresponse = sender.send_image(
                        text_to_addr_formatter(message.to_addr),
                        message.content,
                        image_url=image_url,
                    )
                    log.info("Sent image message to <%s>" % (message.to_addr,))

                else:
                    # Plain content
                    vumiresponse = sender.send_text(
                        text_to_addr_formatter(message.to_addr),
                        message.content,
                        metadata=message.metadata,
                        session_event="new",
                    )
                    log.info("Sent text message to <%s>" % (message.to_addr,))

                message.last_sent_time = timezone.now()
                message.attempts += 1
                message.vumi_message_id = vumiresponse["message_id"]
                message.save()
                fire_metric.apply_async(
                    kwargs={"metric_name": "vumimessage.tries.sum", "metric_value": 1.0}
                )
            except requests_exceptions.ConnectionError as exc:
                log.info("Connection Error sending message")
                fire_metric.delay("sender.send_message.connection_error.sum", 1)
                kwargs["error_retry_count"] = error_retry_count + 1
                self.retry(
                    exc=exc, countdown=retry_delay, args=(message_id,), kwargs=kwargs
                )
            except requests_exceptions.Timeout as exc:
                log.info("Sending message failed due to timeout")
                fire_metric.delay("sender.send_message.timeout.sum", 1)
                kwargs["error_retry_count"] = error_retry_count + 1
                self.retry(
                    exc=exc, countdown=retry_delay, args=(message_id,), kwargs=kwargs
                )
            except (requests_exceptions.HTTPError, HTTPServiceError) as exc:
                # retry message sending if in 500 range (3 default
                # retries)
                log.info(
                    (
                        "Sending message failed due to HTTP response error\n"
                        "status: {response.status_code}\n"
                        "headers: {response.headers}\n"
                        "body: {response.text}\n"
                    ).format(response=exc.response)
                )
                metric_name = (
                    "sender.send_message.http_error.%s.sum" % exc.response.status_code
                )
                fire_metric.delay(metric_name, 1)
                kwargs["error_retry_count"] = error_retry_count + 1
                self.retry(
                    exc=exc, countdown=retry_delay, args=(message_id,), kwargs=kwargs
                )

            # If we've gotten this far the message send was successful.
            fire_metric.apply_async(
                kwargs={"metric_name": "message.sent.sum", "metric_value": 1.0}
            )
            return vumiresponse

        else:
            # This is for retries based on async nacks from the transport.
            log.info("Message <%s> at max retries." % str(message_id))
            message.to_addr = ""
            message.save(update_fields=["to_addr"])
            fire_metric.apply_async(
                kwargs={
                    "metric_name": "vumimessage.maxretries.sum",
                    "metric_value": 1.0,
                }
            )
            # Count failures on exhausted tries.
            fire_metric.apply_async(
                kwargs={"metric_name": "message.failures.sum", "metric_value": 1.0}
            )

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        error_retry_count = kwargs.get("error_retry_count", 0)
        if error_retry_count == self.max_error_retries:
            if "message_id" in kwargs:
                message_id = kwargs["message_id"]
            else:
                message_id = args[0]
            OutboundSendFailure.objects.create(
                outbound_id=message_id,
                initiated_at=self.request.eta,
                reason=einfo.exception.message,
                task_id=task_id,
            )
            # Count permanent failures.
            fire_metric.apply_async(
                kwargs={"metric_name": "message.failures.sum", "metric_value": 1.0}
            )
        super(SendMessage, self).on_failure(exc, task_id, args, kwargs, einfo)


send_message = SendMessage()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed Outbounds.
    """

    name = "message_sender.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        log = self.get_logger(**kwargs)
        failures = OutboundSendFailure.objects
        log.info(
            "Attempting to requeue <%s> failed Outbound sends" % failures.all().count()
        )
        for failure in failures.iterator():
            outbound_id = str(failure.outbound_id)
            # Cleanup the failure before requeueing it.
            failure.delete()
            send_message.delay(outbound_id)


requeue_failed_tasks = RequeueFailedTasks()


class AggregateOutboundMessages(Task):
    """
    Task to aggregate the outbound messages and store the results in the
    aggregate table
    """

    name = "message_sender.tasks.aggregate_outbounds"

    def run(self, start_date, end_date):
        start_date = datetime.strptime(start_date, "%Y-%m-%d").replace(tzinfo=pytz.UTC)
        end_date = datetime.strptime(end_date, "%Y-%m-%d").replace(tzinfo=pytz.UTC)

        # Delete any existing aggregates for these dates. This is necessary
        # to avoid having leftovers from changed objects. eg. There were
        # undelivered messages, but now they're all delivered, so we don't want
        # the undelivered aggregate to still be there, but an update won't set
        # the undelivered aggregate to 0.
        AggregateOutbounds.objects.filter(
            date__gte=start_date.date(), date__lte=end_date.date()
        ).delete()

        for d in daterange(start_date, end_date):
            query = Outbound.objects.filter(
                created_at__gte=d, created_at__lt=(d + timedelta(1))
            )
            query = query.values("delivered", "channel")
            query = query.annotate(attempts=Sum("attempts"), total=Count("*"))
            for aggregate in query.iterator():
                AggregateOutbounds.objects.create(
                    date=d,
                    delivered=aggregate["delivered"],
                    channel_id=aggregate["channel"],
                    attempts=aggregate["attempts"],
                    total=aggregate["total"],
                )


aggregate_outbounds = AggregateOutboundMessages()


class ArchiveOutboundMessages(Task):
    """
    Task to archive the outbound messages and store the messages in the
    storage backend
    """

    name = "message_sender.tasks.archive_outbounds"

    def filename(self, date):
        """
        Returns the filename for the provided date
        """
        return "outbounds-{}.gz".format(date.strftime("%Y-%m-%d"))

    def dump_data(self, filename, queryset):
        """
        Serializes the queryset into a newline separated JSON format, and
        places it into a gzipped file
        """
        with gzip.open(filename, "wb") as f:
            for outbound in queryset.iterator():
                data = OutboundArchiveSerializer(outbound).data
                data = JSONRenderer().render(data)
                f.write(data)
                f.write("\n".encode("utf-8"))

    def create_archived_outbound(self, date, filename):
        """
        Creates the required ArchivedOutbound entry with the file specified
        at `filename`
        """
        with open(filename, "rb") as f:
            f = File(f)
            ArchivedOutbounds.objects.create(date=date, archive=f)

    def run(self, start_date, end_date):
        start_date = datetime.strptime(start_date, "%Y-%m-%d").replace(tzinfo=pytz.UTC)
        end_date = datetime.strptime(end_date, "%Y-%m-%d").replace(tzinfo=pytz.UTC)

        for d in daterange(start_date, end_date):
            if ArchivedOutbounds.objects.filter(date=d.date()).exists():
                continue

            query = Outbound.objects.filter(
                created_at__gte=d, created_at__lt=(d + timedelta(1))
            )

            if not query.exists():
                continue

            filename = self.filename(d)
            self.dump_data(filename, query)
            self.create_archived_outbound(d, filename)

            os.remove(filename)

            # Remove the post_delete hook from rest_hooks, otherwise we'll have
            # to load all of the outbounds into memory
            post_delete.disconnect(
                receiver=model_deleted, dispatch_uid="instance-deleted-hook"
            )
            query.delete()
            post_delete.connect(
                receiver=model_deleted, dispatch_uid="instance-deleted-hook"
            )


archive_outbound = ArchiveOutboundMessages()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-02-20 13:29
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0016_outbound_resend")]

    operations = [
        migrations.AlterField(
            model_name="channel",
            name="channel_type",
            field=models.CharField(
                choices=[
                    (b"junebug", b"Junebug"),
                    (b"vumi", b"Vumi"),
                    (b"http_api", b"HTTP API"),
                ],
                default=b"junebug",
                max_length=20,
            ),
        )
    ]



# Generated by Django 2.1.1 on 2018-09-26 07:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0022_auto_20180914_1426")]

    operations = [
        migrations.AlterField(
            model_name="channel",
            name="channel_type",
            field=models.CharField(
                choices=[
                    ("junebug", "Junebug"),
                    ("vumi", "Vumi"),
                    ("http_api", "HTTP API"),
                    ("wassup", "Wassup API"),
                    ("whatsapp", "WhatsApp API"),
                ],
                default="junebug",
                max_length=20,
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-03-16 13:05
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0017_auto_20180220_1329")]

    operations = [
        migrations.CreateModel(
            name="AggregateOutbounds",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "date",
                    models.DateField(help_text=b"The date that the aggregate is for"),
                ),
                (
                    "delivered",
                    models.BooleanField(
                        help_text=b"Whether this is for delivery passed or failed messages"
                    ),
                ),
                (
                    "attempts",
                    models.IntegerField(help_text=b"The total number of attempts"),
                ),
                (
                    "total",
                    models.IntegerField(help_text=b"The total number of messages"),
                ),
                (
                    "channel",
                    models.ForeignKey(
                        help_text=b"Which channel this is for",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="message_sender.Channel",
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="aggregateoutbounds",
            unique_together=set([("date", "delivered", "channel")]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-11-24 13:57
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0002_auto_20161124_1150")]

    operations = [
        migrations.AlterField(
            model_name="inbound",
            name="transport_type",
            field=models.CharField(blank=True, max_length=200, null=True),
        )
    ]



import codecs
import os
import re

from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):  # Stolen from txacme
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version("seed_message_sender")


setup(
    name="seed-message-sender",
    version=version,
    url="http://github.com/praekelt/seed-message-sender",
    license="BSD",
    description="Seed Message Sender mircoservice",
    long_description=read("README.rst"),
    author="Praekelt.org",
    author_email="dev@praekelt.org",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "Django==2.2.8",
        "djangorestframework==3.9.1",
        "dj-database-url==0.5.0",
        "psycopg2==2.7.5",
        "raven==6.9.0",
        "django-filter==2.0.0",
        "celery==4.2.1",
        "pytz==2018.5",
        "django-rest-hooks==1.5.0",
        "go_http==0.3.2",
        "django-redis==4.10.0",
        "seed-services-client==0.37.0",
        "django-getenv==1.3.2",
        "django-storages==1.7.1",
        "boto3==1.9.4",
        "coreapi==2.3.3",
        "phonenumberslite==8.9.14",
        "django_prometheus==1.0.15",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from django.apps import AppConfig

from django.db.models.signals import post_save, pre_save  # noqa


class MessageSenderAppConfig(AppConfig):

    name = "message_sender"

    def ready(self):
        from .signals import psh_fire_msg_action_if_new, update_default_channels

        post_save.connect(
            psh_fire_msg_action_if_new,
            sender="message_sender.Outbound",
            dispatch_uid="psh_fire_msg_action_if_new",
        )

        post_save.connect(
            update_default_channels,
            sender="message_sender.Channel",
            dispatch_uid="update_default_channels",
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-01-31 10:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0015_identitylookup")]

    operations = [
        migrations.AddField(
            model_name="outbound",
            name="resend",
            field=models.NullBooleanField(
                default=None,
                help_text=b"True if this is a resend requested by the user.",
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-05-08 14:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("message_sender", "0014_auto_20170503_1248")]

    operations = [
        migrations.CreateModel(
            name="IdentityLookup",
            fields=[
                (
                    "msisdn",
                    models.CharField(max_length=255, primary_key=True, serialize=False),
                ),
                ("identity", models.CharField(blank=False, max_length=36, null=False)),
            ],
        )
    ]
